<?php 
/**
 * Plugin name: Hatch
 * Author: A Big Egg
 * Version: 0.4.4
 * Description: Provides helper functions for working with Timber and ACF. Requires Timber (timber-library) and works well with Advanced Custom Fields Pro and Gravity Forms.
 */

require_once( 'inc/image.php' );

register_activation_hook( __FILE__, [ 'Hatch', 'activate_plugin' ] ); 

class Hatch {
	static $context = [];

	static $post_transformers = [];
	static $taxonomy_transformers = [];

	function __construct() {
		// ensure that post objects from ACF are passed through our transformation function
		add_filter( 'acf/format_value/type=post_object', [ $this, 'construct_post' ], 50 );
		add_filter( 'acf/format_value/type=relationship', [ $this, 'construct_posts' ], 50 );
		add_filter( 'acf/format_value/type=image', [ $this, 'construct_image' ], 50 );
	}

	function activate_plugin() {
		update_option( '_abe_hatch_use_timber', true );
	}
	
	/**
	 * Add an ACF field to the Timber array, passing an optional preparation function to prepare the value
	 * 
	 * Here's an example - we can pass in the names of ACF fields we want added to the context:
	 * 
	 * ```php
	 * Hatch::add_acf_context( [ 
	 *    'title',
	 *    'favourite_colour'
	 * ]);
	 * ```
	 * 
	 * Another example, but this time the 'title' field will be passed through our a preparation function to transform it before it's added to the context:
	 * 
	 * ```php
	 * Hatch::add_acf_context( [ 
	 *    'title' => function( $value ) {
	 *        return strtoupper( $value );
	 *     },
	 *     'favourite_colour'
	 * ]);
	 * ```
	 * 
	 * (In any case, you need to call `Hatch::render` to render a template with your context set)
	 *
	 * @param  mixed $keys A string for a single ACF field name, or an array of ACF field names
	 * @return void
	 */
	static function add_acf_context( $key, $func = false ) {
		if ( is_string( $key ) ) {
			$val = get_field( $key );

			if ( is_callable( $func ) ) {
				$val = call_user_func( $func, $val );
			}

			static::$context[$key] = $val;
			return;
		}

		if ( is_array( $key ) ) {
			$keys = $key;
		}

		foreach ( $keys as $key => $value ) {
			if ( is_integer( $key ) && is_string( $value ) ) {
				// we've been passed the name of an ACF field - add it to the context
				$name = $value; // the ACF name is in the value of the array
				static::$context[$name] = get_field( $name );
				continue;
			}

			if ( is_string( $key ) ) {
				// we've been passed the name of an ACF field AND a preparation function
				$func = $value;
				static::$context[$key] = call_user_func( $func, get_field( $key ) );
			}
		}
	}
	
	/**
	 * Add a key/value to the context without touching ACF
	 *
	 * @param  mixed $key The key to add to the context array OR an array of keys
	 * @param  mixed $value The value to set (or a function which will return the value)
	 * @return void
	 */
	static function add_context( $key, $value = false ) {
		if ( is_array( $key ) ) {
			foreach ( $key as $key => $value ) {
				self::add_context( $key, $value );
			}
			return;
		}

		self::$context[$key] = self::get_value_from_thing( $value );
	}

	
	/**
	 * Sometimes it's nice to be able to pass a function OR a different type of variable, for more fluent coding
	 * 
	 * For instance in add_context() we can take either a string or a function as the second value
	 * 
	 * So we could do 
	 *
	 * ```php
	 * add_context( 'key', 'my value' ) 
	 * ```
	 * 
	 * OR
	 * 
	 * ```php
	 * add_context( 'key', function() {
	 *   return 'my value';
	 * } );
	 * ```
	 * 
	 * It's nicer this way!
	 *
	 * @param  mixed $thing
	 * @return void
	 */
	static private function get_value_from_thing( $thing ) {
		if ( is_callable( $thing ) || function_exists( $thing ) ) {
			return call_user_func( $thing );
		}

		return $thing;
	}
	
	/**
	 * Add a Gravity Form to the context
	 *
	 * @param  mixed $key The key to add to the context array
	 * @param  mixed $form_id The ID of the form, or a function which returns the ID
	 * @param  mixed $args The args 
	 * @return void
	 */
	static function add_gform_context( $key, $form_id, $args = [] ) {
		$args = wp_parse_args( $args, [
			'display_title'       => true,
			'display_description' => true,
			'display_inactive'    => false,
			'field_values'        => null,
			'ajax'                => false,
			'tabindex'            => 0,
		] );
		
		$form_id = self::get_value_from_thing( $form_id );

		$rendered_form = gravity_form( $form_id, $args['display_title'], $args['display_description'], $args['display_inactive'], $args['field_values'], $args['ajax'], $args['tabindex'], false );
	
		self::add_context( $key, $rendered_form );
	}
	
	/**
	 * Render a template with Timber using our prepared context
	 *
	 * @param  mixed $view The Timber template to render
	 * @param  mixed $additional_context An optional array of additional context to pass through - this will override the context currently set
	 * @return void
	 */
	static function render( $view, $additional_context = [] ) {
		$context = self::get_final_context( $additional_context );

		Timber::render( $view, $context );

		static::$context = [];
	}
	
	/**
	 * get_final_context
	 *
	 * @param  mixed $context
	 * @return void
	 */
	static private function get_final_context( $context = [] ) {
		$base_context = Timber::get_context();
		$base_context['post'] = self::get_post();
		
		return array_merge( $base_context, static::$context, $context );
	}
	
	/**
	 * Get posts
	 *
	 * @param  mixed $args
	 * @return array
	 */
	static function get_posts( $args ) {
		return self::construct_posts( Timber::get_posts( $args ) );
	}

	/**
	 * get_taxonomy_filter_options
	 *
	 * @param  mixed $args
	 * @return void
	 */
	static function get_taxonomy_filter_options( $args ) {
		$args = wp_parse_args( $args, [
			'taxonomy'   => false,
			'label'      => 'Select an option',
			'filter_key' => 'filter_' . $args['taxonomy'],
			'page'       => false
		] );
		
		$filter_key   = $args['filter_key'];
		$label        = $args['label'];
		$listing_page = $args['listing_page'];
		$active_val   = intval( $_GET[$filter_key] );
		$output = [];

		$terms = self::get_terms( [
			'taxonomy' => $args['taxonomy']
		] );

		array_unshift( $output, [
			'title'  => $label,
			'active' => empty( $active_val ),
			'link'   => add_query_arg( $filter_key, 0, $listing_page )
		] );

		foreach ( $terms as &$term ) {
			$output[] = [
				'title'  => $term->name,
				'value'  => $term->id,
				'active' => $active_val == $term->id,
				'link'   => add_query_arg( $filter_key, $term->id, $listing_page )
			];
		}

		return $output;
	}

	/**
	 * Get terms
	 *
	 * @param  mixed $args
	 * @return array
	 */
	static function get_terms( $args ) {
		return self::construct_terms( Timber::get_terms( $args ) );
	}
	
	/**
	 * get_post
	 *
	 * @return object
	 */
	static function get_post( $post = false ) {
		$the_post = Timber::get_post( $post );

		return self::construct_post( $the_post );
	}
	
	/**
	 * construct_post
	 *
	 * @param  mixed $post_id
	 * @return void
	 */
	static function construct_post( $post_id = 0 ) {
		if ( ! $post_id ) {
			return $post_id;
		}

		$post = new Timber\Post( $post_id );

		$thumb_id = get_post_thumbnail_id( $post );

		if ( $thumb_id ) {
			$post->thumbnail = new HatchImage( $thumb_id );
		}

		if ( self::has_transformer_for_post_type( $post->post_type ) ) {
			$post = self::transform_post( $post );
		}
	

		return $post;
	}
	
		
	/**
	 * construct term
	 *
	 * @param  mixed $term_id
	 * @return void
	 */
	static function construct_term( $term_id = 0 ) {
		if ( ! $term_id ) {
			return $term_id;
		}

		$term = new Timber\Term( $term_id );

		if ( self::has_transformer_for_taxonomy( $term->taxonomy ) ) {
			$term = self::transform_term( $term );
		}
	
		return $term;
	}
	
	
	/**
	 * construct_image
	 *
	 * @param  mixed $image
	 * @return void
	 */
	static function construct_image( $image ) {
		if ( is_array( $image ) ) {
			$image = $image['id'];
		}
		
		return new HatchImage( $image );
	}
	
	/**
	 * has_transformer_for_post_type
	 *
	 * @param  mixed $post_type
	 * @return void
	 */
	static function has_transformer_for_post_type( $post_type ) {
		return isset( self::$post_transformers[$post_type] );
	}
	
	/**
	 * has_transformer_for_term_type
	 *
	 * @param  mixed $taxonomy
	 * @return void
	 */
	static function has_transformer_for_taxonomy( $taxonomy ) {
		return isset( self::$taxonomy_transformers[$taxonomy] );
	}
	

	/**
	 * transform_post
	 *
	 * @param  mixed $post
	 * @return void
	 */
	static function transform_post( $post ) {
		$transformer = self::$post_transformers[$post->post_type];

		if ( ! self::has_transformer_for_post_type( $post->post_type ) ) {
			return $post;
		}

		return call_user_func( $transformer, $post );
	}
	
	/**
	 * transform_term
	 *
	 * @param  mixed $post
	 * @return void
	 */
	static function transform_term( $term ) {
		$transformer = self::$taxonomy_transformers[$term->taxonomy];

		if ( ! self::has_transformer_for_taxonomy( $term->taxonomy ) ) {
			return $term;
		}

		return call_user_func( $transformer, $term );
	}

	/**
	 * construct_post
	 *
	 * @param  mixed $post_id
	 * @return void
	 */
	static function construct_posts( $array_of_posts ) {
		if ( ! is_array( $array_of_posts ) ) {
			return [];
		}

		return array_map( 'self::construct_post', $array_of_posts );
	}

	
	/**
	 * construct_terms
	 *
	 * @param  mixed $post_id
	 * @return void
	 */
	static function construct_terms( $array_of_terms ) {
		if ( ! is_array( $array_of_terms ) ) {
			return [];
		}

		return array_map( 'self::construct_term', $array_of_terms );
	}

	
	
	/**
	 * Register a transformer for a post. A transformer lets you manipulate the Timber post objects
	 * that are returned by Hatch::get_posts or Hatch::get_post
	 *
	 * @param  mixed $post_type
	 * @param  mixed $callback
	 * @return void
	 */
	static function register_transformer_for_post_type( $post_type, $callback ) {
		self::$post_transformers[$post_type] = $callback;
	}
	
	/**
	 * Register a transformer for a taxonomy
	 *
	 * @param  mixed $taxonomy
	 * @param  mixed $callback
	 * @return void
	 */
	static function register_transformer_for_taxonomy( $taxonomy, $callback ) {
		self::$taxonomy_transformers[$taxonomy] = $callback;
	}
	/**
	 * Debug function - call this to dump the current context to inspect it
	 *
	 * @return void
	 */
	static function dd() {
		$context = self::get_final_context();

		if ( function_exists( 'dd' ) ) {
			dd( $context );
		} else {
			var_dump( $context ); die();
		}
	}
}

$h_instance = new Hatch();